require_relative '../wendelin_client'
require 'fileutils'

module Embulk
  module Input

    class Wendelininput < InputPlugin

      CHUNK_SIZE = 50000000 #50mb
      MEGA = 1000000

      UPDATE = "U"
      RESUME = "R"
      DOWNLOAD = "D"
      ABORT = "A"

      TASK_REPORT_FILE = ".download-task-report"
      COMPLETED_FILE = ".download-completed"

      RUN_DONE = "done"
      RUN_ERROR = "error"
      RUN_ABORTED = "aborted"

      Plugin.register_input("wendelin", self)

      def self.getTaskReportFilename(data_set_directory)
	return data_set_directory + TASK_REPORT_FILE
      end

      def self.getCompletedFilename(data_set_directory)
	return data_set_directory + COMPLETED_FILE
      end

      def self.getPendingDataStreams(data_streams, task_report_file)
	donwloaded_data_streams = []
	File.readlines(task_report_file).each do |line|
	  record = line.split(";")
	  if record[1].chomp == RUN_DONE
	    donwloaded_data_streams.push(record[0])
	  end
	end
	pending_data_streams = []
	data_streams.each do |data_stream|
	  if ! donwloaded_data_streams.include? data_stream[1]
	    pending_data_streams.push(data_stream)
	  end
	end
	return pending_data_streams
      end

      def self.askUserForAction(task, completed_file, task_report_file, action)
	    if action == RESUME
	      action_message =  "#{RESUME}: Resume. Continues download from last file."
	    else
	      action = UPDATE
	      action_message = "#{UPDATE}: Update. Checks for new files."
            end
	    valid_option = FALSE
	    while not valid_option
	      @logger.info("Please select an option [#{action}, #{DOWNLOAD}, #{ABORT}]", print=TRUE)
	      @logger.info(action_message, print=TRUE)
	      @logger.info("#{DOWNLOAD}: Download. Downloads the dataset from scratch.", print=TRUE)
	      @logger.info("#{ABORT}: Abort operation.", print=TRUE)
	      option = gets
	      option = option.chomp
	      if not [action, DOWNLOAD, ABORT].include? option
		@logger.info("Invalid option", print=TRUE)
	      else
		valid_option = TRUE
	      end
	    end
	    case option
	      when action
		task['data_streams'] = getPendingDataStreams(task['data_streams'], task_report_file)
		File.delete(completed_file) if File.exist?(completed_file)
	        if task['data_streams'].empty?
	          @logger.info("No new files in dataset.", print=TRUE)
	          @logger.info("Your downloaded dataset is already up to date.", print=TRUE)
	        end
	      when DOWNLOAD
		ebulk_file = @data_set_directory + "/.ebulk_dataset"
		ebulk_file_content = ""
		if File.file?(ebulk_file)
		  ebulk_file_content = File.read(ebulk_file)
		end
		FileUtils.rm_rf(@data_set_directory)
		unless File.directory?(@data_set_directory)
		  FileUtils.mkdir_p(@data_set_directory)
		end
		if ebulk_file_content != ""
		  File.open(ebulk_file, 'w') { |file| file.write(ebulk_file_content) }
		end
	        File.delete(completed_file) if File.exist?(completed_file)
		File.open(task_report_file, 'w') {}
	      when ABORT
		@logger.abortExecution()
	    end
      end

      def self.transaction(config, &control)
	begin
		@tool_dir = config.param('tool_dir', :string)
		@logger = LogManager.instance()
		@logger.setFilename(@tool_dir, "download")
		@erp5_url = config.param('erp5_url', :string)
		@data_set = config.param('data_set', :string)
		@logger.info("Dataset name: #{@data_set}")
		if @data_set == '$DATA_SET'
		  @logger.error("There was an error setting the configuration file", print=TRUE)
		  @logger.info("Please try manual download or update manually the download configuration file.", print=TRUE)
		  @logger.abortExecution()
		end
		@user = config.param("user", :string, defualt: nil)
		@logger.info("User: #{@user}")
		@password = config.param("password", :string, default: nil)
		@chunk_size = config.param('chunk_size', :float, default: 0) * MEGA
		@output_path = config.param("output_path", :string,  :default => nil)
		if File.directory?(@output_path)
		else
		  @logger.error("Output directory not found.", print=TRUE)
		  @logger.abortExecution()
		end
		@wendelin = WendelinClient.new(@erp5_url, @user, @password)

		task = { 
			 'erp5_url' => @erp5_url,
			 'data_set' => @data_set,
		  	 'user' => @user,
		  	 'password' => @password,
			 'chunk_size' => @chunk_size,
			 'output_path' => @output_path,
			 'tool_dir' => @tool_dir
		       }
		if task['chunk_size'] == 0
		  task['chunk_size'] = CHUNK_SIZE
		end
		@logger.info("Chunk size set in #{task['chunk_size']/MEGA}MB")
		@data_set_directory = @output_path.end_with?("/") ? @output_path : @output_path + "/"
		task['data_set_directory'] = @data_set_directory

		data_stream_list = @wendelin.getDataStreams(@data_set)
		if data_stream_list["status_code"] == 0
		  if data_stream_list["result"].empty?
		    @logger.error("No valid data found for data set " + @data_set, print=TRUE)
		    @logger.info("Please use a valid dataset reference from the list of datasets available in the site.", print=TRUE)
		    @logger.abortExecution()
		  end
		  task['data_streams'] = data_stream_list["result"]
		else
		  @logger.error(data_stream_list["error_message"], print=TRUE)
		  @logger.abortExecution()
		end

		task_report_file = getTaskReportFilename(@data_set_directory)
		completed_file = getCompletedFilename(@data_set_directory)
		if File.file?(task_report_file)
		  if File.file?(completed_file)
		    puts
		    @logger.info("This dataset was already downloaded. What do you want to do?", print=TRUE)
		    puts
		    self.askUserForAction(task, completed_file, task_report_file, action=UPDATE)
		  else
		    puts
		    @logger.info("There was a previous attempt to download this dataset but it did not finish successfully.", print=TRUE)
		    @logger.info("What do you want to do?", print=TRUE)
		    puts
		    self.askUserForAction(task, completed_file, task_report_file, action=RESUME)
		  end
		else
		  dir_entries = Dir.entries(@data_set_directory).length
		  if File.file?(@data_set_directory+"/.ebulk_dataset")
		    dir_entries -= 1
		  end
		  if dir_entries > 2
		    puts
		    @logger.info("Dataset download directory is not empty! It will be overwritten: " + @data_set_directory, print=TRUE)
		    @logger.info("Continue with download? (y/n)", print=TRUE)
		    option = gets
		    option = option.chomp
		    if option == "n"
		      @logger.info("Download cancelled by user.", print=TRUE)
		      @logger.abortExecution()
		    end
		  end
		  ebulk_file = @data_set_directory + "/.ebulk_dataset"
		  ebulk_file_content = ""
		  if File.file?(ebulk_file)
		    ebulk_file_content = File.read(ebulk_file)
		  end
		  FileUtils.rm_rf(@data_set_directory)
		  unless File.directory?(@data_set_directory)
		    FileUtils.mkdir_p(@data_set_directory)
		  end
		  if ebulk_file_content != ""
		    File.open(ebulk_file, 'w') { |file| file.write(ebulk_file_content) }
		  end
		  File.open(task_report_file, 'w') {}
		end
		columns = [
		  Column.new(0, "reference", :string),
		  Column.new(1, "data_chunk", :string),
		  Column.new(2, "data_set", :string)
		]
		resume(task, columns, task['data_streams'].length, &control)
	rescue Exception => e
	  @logger.error("An error occurred during operation: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file."
	  @logger.abortExecution()
	end
      end

      def self.resume(task, columns, count, &control)
	@logger = LogManager.instance()
        task_reports = yield(task, columns, count)
	if task_reports.any?
	  @logger.info("Reports:", print=TRUE)
	  if task_reports.length > 15
	    @logger.info(task_reports[0, 5], print=TRUE)
	    @logger.info(".....", print=TRUE)
	    @logger.info(task_reports[task_reports.length-5, task_reports.length-1], print=TRUE)
	  else
	    @logger.info(task_reports, print=TRUE)
	  end
	  @logger.info("Full task report:")
	  @logger.info(task_reports)
	end
        next_config_diff = task_reports.map{|hash| hash[RUN_DONE]}.flatten.compact
	if(next_config_diff.length == count)
	  @logger.info("Dataset successfully downloaded.", print=TRUE)
	  @logger.info("#{count} files downloaded.", print=TRUE)
	  @logger.info("The files were saved in dataset directory: " + @data_set_directory, print=TRUE)
	  completed_file = getCompletedFilename(@data_set_directory)
	  File.open(completed_file, 'w') {}
	  if count > 10
	    next_config_diff = {}
	  end
	end
	return {RUN_DONE => next_config_diff}
      end

      def initialize(task, schema, index, page_builder)
        super
        @data_set = task['data_set']
	@chunk_size = task['chunk_size']
	@data_set_directory = task['data_set_directory']
	@wendelin = WendelinClient.new(task['erp5_url'], task['user'], task['password'])
	@logger = LogManager.instance()
      end

      def run
	data_stream = task['data_streams'][@index]
	id = data_stream[0]
	ref = data_stream[1]
	@logger.info("Getting content from remote #{ref}", print=TRUE)
	begin
  	  @wendelin.eachDataStreamContentChunk(id, @chunk_size) do |chunk|
	    if chunk.nil? || chunk.empty?
	      content = ""
	    else
	      content = Base64.encode64(chunk)
	    end
	    entry = [ref, content, @data_set]
  	    page_builder.add(entry)
          page_builder.finish
	  end
	rescue java.lang.OutOfMemoryError
          @logger.logOutOfMemoryError(ref)
	  return_value = RUN_ABORTED
        rescue Exception => e  
	  @logger.error("An error occurred during processing: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file: " + @logger.getLogPath()
	  return_value = RUN_ERROR
	else
	  return_value = RUN_DONE
	end
	task_report_file = Wendelininput.getTaskReportFilename(@data_set_directory)
	File.open(task_report_file, 'ab') { |file| file.puts(ref+";"+return_value+";") }
        return {return_value => ref}
      end
    end
  end
end
